import os
import subprocess
import json
import uuid
import time
import urllib.parse
from jinja2 import Environment, FileSystemLoader
from flask import Flask, send_file, make_response, request, copy_current_request_context
from threading import Thread
from flask_restplus import Resource, Api, fields

app = Flask(__name__)
app.config.SWAGGER_UI_JSONEDITOR = True
app.config.SWAGGER_UI_DOC_EXPANSION = 'list'

api = Api(app, doc='/doc/')

env_fields = api.model('EnvModel', {
    'cloudbroker_endpoint': fields.String(description='The Cloudbroker REST Api endpoint', required=True),
    'cloudbroker_username': fields.String(description='The username', required=True),
    'cloudbroker_password': fields.String(description='The password', required=True),
    'repository_user': fields.String(description='The Git repo username', required=True),
    'repository_pwd': fields.String(description='The Git repo password', required=True),
    'repository_url': fields.String(description='The Git repo http(s) url', required=True),
    'revision': fields.String(description='The code revision for the configuration', required=True),
    'filepath': fields.String(description='The filepath for the configuration file in the repo', required=True),
})

@app.route("/")
def main():
    index_path = os.path.join(app.static_folder, 'index.html')
    return send_file(index_path)


@api.route("/terraform/<environment>")
@api.doc(params={'environment': 'The environment [infrastructure | services] for which we want to create the configuration files'})
class Configuration(Resource):
    @api.expect(env_fields)
    def post(self, environment):
        data = {}
        body = request.get_json()
        print(body)
        if not "repository_url" in body or not "revision" in body or not "filepath" in body or not body["repository_user"] or not body["repository_pwd"]:
            data["result"] = "error"
            data["resultCode"] = "400"
            data["content"] = "Missing parameters"
            response = make_response(json.dumps(data))
            response.status_code = 400
            return response
        repository = body["repository_url"].split('/')[-1]
        repository = repository[:repository.index('.git')]
        revision = body["revision"]
        filePath = body["filepath"]
        if not os.path.exists("repo/"+repository):
            print("Clone repository: " + body["repository_url"])
            repo_user = body["repository_user"]
            repo_pwd = body["repository_pwd"]
            repo_url = body["repository_url"]
            repo_url = repo_url[8:]
            repo_url = "https://"+urllib.parse.quote(repo_user)+":"+urllib.parse.quote(repo_pwd)+"@"+repo_url
            print("repo_url for clone: " + repo_url)
            result = subprocess.check_output(['git', 'clone', repo_url], cwd="repo/")
        print("Getting revision " + revision + " for file " + filePath + " from repository " + repository + " (" + body["repository_url"] + ")")
        result = subprocess.check_output(['git', 'pull'], cwd="repo/"+repository)
        result = subprocess.check_output(['git', 'checkout', '-m', revision, filePath], cwd="repo/"+repository)
        file = open('repo/'+repository+"/"+filePath, 'r') 
        content = file.read() 
    
        print("Creating config for " + environment + " with input file:")
        print(json.loads(content))
        jsonContent = json.loads(content)
        if "cloudbroker_username" in body and "cloudbroker_password" in body and "cloudbroker_endpoint" in body and body["cloudbroker_username"] and body["cloudbroker_password"] and body["cloudbroker_endpoint"]:
            jsonContent["cloudbrokerEndpoint"] = body["cloudbroker_endpoint"]
            jsonContent["cloudbrokerUsername"] = body["cloudbroker_username"]
            jsonContent["cloudbrokerPassword"] = body["cloudbroker_password"]
            print("Changed credentials!")

        data["result"] = "success"
        data["resultCode"] = "202"

        if environment == "infrastructure":
            create_infrastructure_tf(jsonContent)
            data["content"] = "Request for infrastructure configuration sent"
            return make_response(json.dumps(data))
        elif environment == "services":
            create_services_tf(jsonContent)
            data["content"] = "Request for services configuration sent"
            return make_response(json.dumps(data))
        elif environment == "all":
            create_infrastructure_tf(jsonContent)
            create_services_tf(jsonContent)
            data["content"] = "Request for infrastructure and services configuration sent"
            return make_response(json.dumps(data))
        else:
            data["result"] = "error"
            data["resultCode"] = "400"
            data["content"] = "Wrong environment parameter. Accepted values are: 'infrastructure'|'services'|'all'"
            response = make_response(json.dumps(data))
            response.status_code = 400
            return response


@api.route("/terraform/<operation>/<appName>/<folder>")
@api.doc(params={'operation': 'The Terraform operation to perform: [init | plan | apply]', 'appName':'The name of the application','folder':'The logical application area we want to manage: [infrastructure | services]'})
class terraformOperation(Resource):
    def post(self, operation, appName, folder):
        op_id = str(uuid.uuid4())
    
        @copy_current_request_context
        def terraformThread(operation, appName, folder, op_id):
            print("Starting thread for operation " + op_id)
            log_dir = appName+"/logs"
            if not os.path.exists(log_dir):
                os.makedirs(log_dir)
            ops_dir = "/app/operations"
            if not os.path.exists(ops_dir):
                os.makedirs(ops_dir)
            with open(ops_dir+"/"+op_id+".txt", 'w') as f:
                f.write("running")
    
            #result = make_response(subprocess.run(['terraform', operation, '-no-color'], cwd=appName+"/"+folder, stdout=subprocess.PIPE, stderr=subprocess.STDOUT).stdout.decode('utf-8'))
            with open(log_dir+"/"+folder+"-"+operation+"-"+op_id+".log","w") as out:
                if operation == "destroy":
                    myproc = subprocess.run(['terraform', operation, '-no-color', '-force'], cwd=appName+"/"+folder, stdout=out, stderr=out)
                else:
                    print("Starting subprocess")
                    myproc = subprocess.run(['terraform', operation, '-no-color'], cwd=appName+"/"+folder, stdout=out, stderr=out)
            print("myproc return code: " + str(myproc.returncode))
            with open(ops_dir+"/"+op_id+".txt", 'w') as f:
                f.write(str(myproc.returncode))
    
        t = Thread(target=terraformThread, args=(operation, appName, folder, op_id))
        t.start()
        data = {}
        data["operation_id"] = op_id 
        data["get_log_url"] = request.url_root + "terraform/"+operation+"/"+appName+"/"+folder+"/"+op_id 
        data["get_result_url"] = request.url_root + "terraform/operation/result/"+op_id 
        response = make_response(json.dumps(data))
        response.headers['Content-Type'] = 'application/json'
        response.status_code = 202
        return response


@api.route("/terraform/<operation>/<appName>/<folder>/<op_id>")
@api.doc(params={'op_id': 'The Operation ID returned by the POSTed operation request'})
class OperationLog(Resource):
    @api.doc(responses={200: 'Ok'})
    def get(self, operation, appName, folder, op_id):
        print("Reading from file: " + appName+"/logs/"+folder+"-"+operation+"-"+op_id+".log")
        with open(appName+"/logs/"+folder+"-"+operation+"-"+op_id+".log", 'r') as f:
            logdata = f.read()
            if len(logdata) > 0:
                result = "0"
            else:
                result = "-1"

        #result = subprocess.run(['tail', appName+"/logs/"+folder+"-"+operation+"-"+op_id+".log"], stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        data = {}
        if result == "0":
            data["result"] = "success"
        else:
            data["result"] = "failure"
        data["resultCode"] = result
        print("Output: " + str(logdata))
        data["content"] = logdata
        response = make_response(json.dumps(data))
        response.headers['Content-Type'] = 'application/json'
        response.status_code = 200
        return response

@api.route("/terraform/operation/result/<op_id>")
@api.doc(params={'op_id': 'The Operation ID returned by the POSTed operation request'})
class OperationStatus(Resource):
    def get(self, op_id):
        print("Reading from file: /app/operations/"+op_id+".txt")
        with open("/app/operations/"+op_id+".txt", 'r') as f:
            lines = f.read().splitlines()
            if len(lines) > 0:
                result = lines[0]
            else:
                result = "-1" 

        data = {}
        print("result: " + result)
        if result == "0":
            data["result"] = "success"
        elif result == "running":
            data["result"] = "running"
        else: 
            data["result"] = "failure"
        data["resultCode"] = result
        response = make_response(json.dumps(data))
        response.headers['Content-Type'] = 'application/json'
        response.status_code = 200
        return response

#@app.route("/terraform/<operation>/<appName>/<folder>/<op_id>", methods=['GET'])
#def getCmdLog(operation, appName, folder, op_id):
#    print("Reading from file: " + appName+"/logs/"+folder+"-"+operation+"-"+op_id+".log")
#    result = subprocess.run(['tail', appName+"/logs/"+folder+"-"+operation+"-"+op_id+".log"], stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
#    response = make_response(result.stdout)
#    response.status_code = 200
#    return response
#
#@app.route("/terraform/operation/result/<op_id>", methods=['GET'])
#def getCmdStatus(op_id):
#    print("Reading from file: /app/operations/"+op_id+".txt")
#    with open("/app/operations/"+op_id+".txt", 'r') as f:
#        result = f.readline() 
#
#    response = make_response(result)
#    response.status_code = 200
#    return response


@api.route("/terraform/adapt/prepare")
class adaptDeploymentPreparation(Resource):
    def post(self):
        jsonContent = request.get_json()
        create_adapt_tf(jsonContent)
        return make_response("Request for 'adapt' configuration sent.\n")


@app.route("/sleep/<seconds>")
def test(seconds):
    subprocess.Popen(['whoami'])
    subprocess.run(['sleep', seconds])
    result = make_response("I went to sleep...")
    result.headers['Content-Type'] = 'text/plain'
    return result


# Everything not declared before (not a Flask route / API endpoint)...
@app.route('/<path:path>')
def route_frontend(path):
    # ...could be a static file needed by the front end that
    # doesn't use the `static` path (like in `<script src="bundle.js">`)
    file_path = os.path.join(app.static_folder, path)
    if os.path.isfile(file_path):
        return send_file(file_path)
    # ...or should be handled by the SPA's "router" in front end
    else:
        index_path = os.path.join(app.static_folder, 'index.html')
        return send_file(index_path)


########################
PATH = os.path.dirname(os.path.abspath(__file__))
TEMPLATE_ENVIRONMENT = Environment(
    autoescape=False,
    loader=FileSystemLoader(os.path.join(PATH, 'templates')),
    trim_blocks=True)


def render_template(template_filename, context):
    return TEMPLATE_ENVIRONMENT.get_template(template_filename).render(context)


def create_infrastructure_tf(data_loaded):

    appName = data_loaded["name"]
    infrastructure_dir = appName+"/infrastructure"
    service_dir = appName+"/services"
    if not os.path.exists(appName):
        os.makedirs(appName)
    if not os.path.exists(infrastructure_dir):
        os.makedirs(infrastructure_dir)
    if not os.path.exists(service_dir):
        os.makedirs(service_dir)

    cloudbrokerEndpoint = data_loaded["cloudbrokerEndpoint"]
    cloudbrokerUsername = data_loaded["cloudbrokerUsername"]
    cloudbrokerPassword = data_loaded["cloudbrokerPassword"]

    for element in data_loaded["virtualMachines"]:
        print(element)
        fname = infrastructure_dir+"/"+appName+"-vm-"+element["dockerHostNodeName"]+".tf"
        context = {
                'appName': appName,
                #'docker_host_ip': element["docker-host-ip"],
                'dockerHostNodeName': element["dockerHostNodeName"],
                'cloudbrokerEndpoint': cloudbrokerEndpoint,
                'cloudbrokerUsername': cloudbrokerUsername,
                'cloudbrokerPassword': cloudbrokerPassword,
                'vmSoftwareId': element["vmSoftwareId"],
                'vmResourceId': element["vmResourceId"],
                'vmRegionId': element["vmRegionId"],
                'instanceTypeId': element["instanceTypeId"],
                'keyPairId': element["keyPairId"],
                'openedPort': element["openedPort"],
                'consulJoinIp': element["consulJoinIp"],
        }
        if 'dockerPrivateRegistryIp' in element:
               context['dockerPrivateRegistryIp'] = element["dockerPrivateRegistryIp"]
        if 'dockerPrivateRegistryPort' in element:
                context['dockerPrivateRegistryPort'] =  element["dockerPrivateRegistryPort"]

        with open(fname, 'w') as f:
                tf = render_template('vm-tpl.tf', context)
                f.write(tf)




def get_docker_host_node_name(container_name, containers):
    for container in containers:
        if container["containerName"] == container_name:
                return container["dockerHostNodeName"]

    return ""

def create_services_tf(data_loaded):

    appName = data_loaded["name"]

    infrastructure_dir = appName+"/infrastructure"
    services_dir = appName+"/services"
    if not os.path.exists(appName):
        os.makedirs(appName)
    if not os.path.exists(infrastructure_dir):
        os.makedirs(infrastructure_dir)
    if not os.path.exists(services_dir):
        os.makedirs(services_dir)

    networks_to_create = set()
    docker_providers = set()
    for element in data_loaded["containers"]:
        print(element)
        fname = services_dir+"/"+appName+"-container-"+element["containerName"]+".tf"
        dockerHostNodeName = element["dockerHostNodeName"]
        context = {
                'appName': appName,
                'containerName': element["containerName"],
                'imageName': element["imageName"],
                #'docker_host_ip': element["docker-host-ip"],
                'dockerHostNodeName': element["dockerHostNodeName"],
                'hostname': element["hostname"],
        }
        docker_providers.add(element["dockerHostNodeName"])
        if 'imageTag' in element:
               context['imageTag'] = element["imageTag"]
        else:
               context['imageTag'] = "latest"
        if 'dockerPrivateRegistryIp' in element:
               context['dockerPrivateRegistryIp'] = element["dockerPrivateRegistryIp"]
        if 'dockerPrivateRegistryPort' in element:
               context['dockerPrivateRegistryPort'] =  element["dockerPrivateRegistryPort"]
        if 'dockerPrivateRegistryUser' in element:
               context['dockerPrivateRegistryUser'] = element["dockerPrivateRegistryUser"]
        if 'dockerPrivateRegistryPassword' in element:
               context['dockerPrivateRegistryPassword'] = element["dockerPrivateRegistryPassword"]
        if 'command' in element:
               context['command'] = json.dumps(element["command"])
        if 'entrypoint' in element:
               context['entrypoint'] = json.dumps(element["entrypoint"])
        if 'networks' in element:
               networks_dependencies = set()
               context['networks'] = element["networks"]
               for network in element["networks"]:
                        networks_dependencies.add("docker_network."+dockerHostNodeName+"-"+network)
                        if dockerHostNodeName+"-"+network not in networks_to_create:
                                networks_to_create.add(dockerHostNodeName+":"+network)
               context['networksDependencies'] = networks_dependencies
        if 'restart' in element:
               context['restart'] = element["restart"]
        if 'environment' in element:
               context['environment'] = json.dumps(element["environment"])
        if 'portMapping' in element:
               context['portMapping'] = element["portMapping"]
        if 'volumeMapping' in element:
               context['volumeMapping'] = element["volumeMapping"]
        if 'hostMapping' in element:
               for hostMappingElement in element["hostMapping"]:
                        if hostMappingElement['referencedContainerName']:
                                hostMappingElement['dockerHostNodeName'] = get_docker_host_node_name(hostMappingElement['referencedContainerName'], data_loaded["containers"])
               context['hostMapping'] = element["hostMapping"]
        if 'endpoints' in element:
               context['endpoints'] = element["endpoints"]
        if 'capabilities' in element:
               context['capabilities'] = element["capabilities"]
        if 'addConsulService' in element:
               context['addConsulService'] = element["addConsulService"]
        if 'consulServicePort' in element:
               context['consulServicePort'] = element["consulServicePort"]
        if 'addConsulTraefikRules' in element:
               context['addConsulTraefikRules'] = element["addConsulTraefikRules"]
        with open(fname, 'w') as f:
                tf = render_template('container-tpl.tf', context)
                f.write(tf)

    #Create common stuff section:
    context = {
        'appName': appName,
        'dockerProviders': docker_providers
    }

    fname = services_dir+"/"+appName+"-services-common.tf"
    with open(fname, 'w') as f:
        tf = render_template('services-common-tpl.tf', context)
        f.write(tf)

    #Create networks section:
    for network_to_create in networks_to_create:
        i = network_to_create.find(':')
        node_name =  network_to_create[:i]
        network_name = network_to_create[i+1:]
        fname = services_dir+"/"+appName+"-network-"+node_name+"-"+network_name+".tf"
        context = {
                'dockerHostNodeName': node_name,
                'networkName': network_name
        }

        with open(fname, 'w') as f:
                tf = render_template('network-tpl.tf', context)
                f.write(tf)


def create_adapt_tf(data_loaded):

    appName = data_loaded["name"]
    adapt_dir = appName+"-adapt"
    infrastructure_dir = adapt_dir+"/infrastructure"
    services_dir = adapt_dir+"/services"
    if not os.path.exists(appName):
        os.makedirs(appName)
    if not os.path.exists(adapt_dir):
        os.makedirs(adapt_dir)
    if not os.path.exists(infrastructure_dir):
        os.makedirs(infrastructure_dir)
    if not os.path.exists(services_dir):
        os.makedirs(services_dir)

    cloudbrokerEndpoint = data_loaded["cloudbrokerEndpoint"]
    cloudbrokerUsername = data_loaded["cloudbrokerUsername"]
    cloudbrokerPassword = data_loaded["cloudbrokerPassword"]

    for element in data_loaded["virtualMachines"]:
        fname = infrastructure_dir+"/"+appName+"-vm-"+element["dockerHostNodeName"]+".tf"
        print(element)
        context = {
                'appName': appName,
                'dockerHostNodeName': element["dockerHostNodeName"],
                'cloudbrokerEndpoint': cloudbrokerEndpoint,
                'cloudbrokerUsername': cloudbrokerUsername,
                'cloudbrokerPassword': cloudbrokerPassword,
                'vmSoftwareId': element["vmSoftwareId"],
                'vmResourceId': element["vmResourceId"],
                'vmRegionId': element["vmRegionId"],
                'instanceTypeId': element["instanceTypeId"],
                'keyPairId': element["keyPairId"],
                'openedPort': element["openedPort"],
                'consulJoinIp': element["consulJoinIp"]
        }
        if 'dockerPrivateRegistryIp' in element:
               context['dockerPrivateRegistryIp'] = element["dockerPrivateRegistryIp"]
        if 'dockerPrivateRegistryPort' in element:
                context['dockerPrivateRegistryPort'] =  element["dockerPrivateRegistryPort"]
        with open(fname, 'w') as f:
            tf = render_template('adapt-vm-tpl.tf', context)
            f.write(tf)


    docker_providers = set()
    for element in data_loaded["containers"]:
        fname = services_dir+"/"+appName+"-container-"+element["dockerHostNodeName"]+".tf"
        print(element)
        dockerHostNodeName = element["dockerHostNodeName"]
        context['containerName'] = element["containerName"]
        context['imageName'] = element["imageName"]
        context['dockerHostNodeName'] = element["dockerHostNodeName"]
        context['hostname'] = element["hostname"]
        docker_providers.add(element["dockerHostNodeName"])
        if 'imageTag' in element:
               context['imageTag'] = element["imageTag"]
        else:
               context['imageTag'] = "latest"
        if 'dockerPrivateRegistryUser' in element:
               context['dockerPrivateRegistryUser'] = element["dockerPrivateRegistryUser"]
        if 'dockerPrivateRegistryPassword' in element:
               context['dockerPrivateRegistryPassword'] = element["dockerPrivateRegistryPassword"]
        if 'command' in element:
               context['command'] = json.dumps(element["command"])
        if 'entrypoint' in element:
               context['entrypoint'] = json.dumps(element["entrypoint"])
        if 'restart' in element:
               context['restart'] = element["restart"]
        if 'environment' in element:
               context['environment'] = json.dumps(element["environment"])
        if 'portMapping' in element:
               context['portMapping'] = element["portMapping"]
        if 'hostMapping' in element:
               for hostMappingElement in element["hostMapping"]:
                        if hostMappingElement['referencedContainerName']:
                                hostMappingElement['dockerHostNodeName'] = get_docker_host_node_name(hostMappingElement['referencedContainerName'], data_loaded["containers"])
               context['hostMapping'] = element["hostMapping"]
        if 'endpoints' in element:
               context['endpoints'] = element["endpoints"]
        if 'capabilities' in element:
               context['capabilities'] = element["capabilities"]
        
        with open(fname, 'w') as f:
            tf = render_template('adapt-container-tpl.tf', context)
            f.write(tf)




########################





if __name__ == "__main__":
    # Only for debugging while developing
    app.run(host='0.0.0.0', debug=True, port=80)

